/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: io.c,v 1.20 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.15
 * Notes:
 *    - at messages: no process identifier!
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.15
 * Notes:
 *    - at messages: no process identifier!
 *
 ***************************************************************************/

#include <zorp/io.h>
#include <zorp/log.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>

/*+ 

  ZIORealWatchFD is an implementation of the ZIOWatchFD interface. It
  uses standard glib hook mechanisms, and g_main loop, and calls its
  registered callback upon receipt of the given event.

  +*/
typedef struct _ZIORealWatchFD
{
  GIOCondition cond;
  GPollFD pollfd;
  ZWatchFDFunc callback;
} ZIORealWatchFD;

/*+ A callback registered as the prepare function of our GSource. +*/
static gboolean 
z_io_watch_fd_prepare(gpointer source,
		      GTimeVal *current_time,
		      gint *timeout,
		      gpointer user_data)
{
  ZIORealWatchFD *fd = (ZIORealWatchFD *) source;

#if 0
  z_debug(10, "watch_prepare, fd=%d, cond=%x", fd->pollfd.fd, fd->cond);
#endif
  fd->pollfd.events = fd->cond;
  fd->pollfd.revents = 0;
  *timeout = -1;
  return FALSE;
}

/*+ Returns TRUE if our source has some data available. +*/
static gboolean
z_io_watch_fd_check(gpointer source_data,
		    GTimeVal *current_time,
		    gpointer user_data)
{
  ZIORealWatchFD *fd = (ZIORealWatchFD *) source_data;

  return !!fd->pollfd.revents;
}

/*+ Called by glib if our source has some data available to dispatch the pending event. +*/
static gboolean
z_io_watch_fd_dispatch(gpointer source_data,
                          GTimeVal *current_time,
                          gpointer user_data)
{
  ZIORealWatchFD *watch = (ZIORealWatchFD *) source_data;
  
  return (*watch->callback)(watch->pollfd.fd, watch->pollfd.revents, user_data);
}

/*+ Called by glib to destroy a source instance. +*/
static void
z_io_watch_fd_destroy_source(gpointer source_data)
{
  ZIORealWatchFD *watch = (ZIORealWatchFD *) source_data;

  g_main_remove_poll(&watch->pollfd);
  g_free(source_data);
}

/*+ Source functions associated with our data source. +*/
GSourceFuncs watch_fd_funcs = 
{
  z_io_watch_fd_prepare,
  z_io_watch_fd_check,
  z_io_watch_fd_dispatch,
  z_io_watch_fd_destroy_source
};


/*+
  
  This function creates a new instance of ZIOWatchFD, capable of
  monitoring the given file descriptor. Upon the occurance of the
  specified I/O condition, the passed callback function is called.

  Parameters:
     fd         fd to monitor
     cond       cond to look for
     callback   function to call
     user_data  opaque pointer to pass to callback
  +*/
ZIOWatchFD *
z_io_watch_fd_new(int fd, 
		  GIOCondition cond,
		  ZWatchFDFunc callback, 
		  gpointer user_data)
{
  ZIORealWatchFD *watch = g_new(ZIORealWatchFD, 1);
  
  watch->pollfd.fd = fd;
  watch->pollfd.events = cond;
  watch->callback = callback;
  watch->cond = cond;
  
  g_main_add_poll(&watch->pollfd, 0);
  g_source_add(G_PRIORITY_DEFAULT, 0, &watch_fd_funcs, watch, user_data, NULL);
  return (ZIOWatchFD *) watch;
}

/*+ Destroy the given ZIOWatchFD instance +*/
void 
z_io_watch_fd_destroy(ZIOWatchFD *watch)
{
  g_source_remove_by_source_data(watch);
}

/*+
  
  The ZIOConnect interface can be used to connect a socket to a given
  destination, using the given local address. The connection establishment
  is started asynchronously, and a callback is invoked when upon completition.
  When the operation is completed this instance is automatically freed.

  +*/
struct _ZIOConnect 
{
  ZSockAddr *local;
  ZSockAddr *remote;
  ZIOWatchFD *watch;
  int sock;
  ZConnectFunc callback;
  gpointer user_data;
  GMutex *lock;
};

/*+ 

  Private callback function, registered to a ZIOWatchFD when the
  socket becomes writeable, e.g. when the connection is established.

  +*/
static gboolean z_io_connect_connected(int fd,
                                       GIOCondition condition,
                                       gpointer data)
{
  ZIOConnect *self = (ZIOConnect *) data;
  ZConnectFunc callback = self->callback;
  gpointer user_data = self->user_data;
  int error = 0;
  socklen_t errorlen = sizeof(error);
  
  g_mutex_lock(self->lock);
  if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &errorlen) == -1)
    {
      z_warning(3, "getsockopt() failed due to: %m, ignoring.");
    }
  if (error)
    {
      /* connect failed */
      z_warning(0, "Connection failed due to: %s.", g_strerror(error));
      close(fd);
      fd = -1;
    }
  else
    {
      z_fd_set_nonblock(fd, 0);
    }
  self->sock = -1;
  g_mutex_unlock(self->lock);
  z_io_connect_destroy(self);
  callback(fd, user_data);
  
  return FALSE;
}

/*+
  
  Public function to create a ZIOConnect instance.

  Parameters:
    local           local address to bind to
    remote          remote address to connect to
    callback        function to call when the connection is established
    user_data       opaque pointer to pass to callback

  Returns:
    The allocated instance.

  +*/
ZIOConnect *z_io_connect_new(ZSockAddr *local, 
			     ZSockAddr *remote,
			     ZConnectFunc callback,
			     gpointer user_data)
{
  ZIOConnect *self = g_new0(ZIOConnect, 1);

  self->local = z_sockaddr_ref(local);
  self->remote = z_sockaddr_ref(remote);
  self->callback = callback;
  self->user_data = user_data;
  self->lock = g_mutex_new();
  self->sock = socket(remote->sa.sa_family, SOCK_STREAM, 0);
  if (self->sock < 0)
    {
      z_message(0, "creating socket failed due to: %s.", g_strerror(errno));
      z_io_connect_destroy(self);
      return NULL;
    }
  if (local && z_bind(self->sock, local) != G_IO_ERROR_NONE)
    {
      z_io_connect_destroy(self);
      return NULL;
    }
  if (!z_fd_set_nonblock(self->sock, TRUE))
    {
      z_message(0, "setting fd %d to nonblocking mode failed due to: %s.", self->sock, g_strerror(errno));
      z_io_connect_destroy(self);
      return NULL;
    }
  if (z_connect(self->sock, remote) != G_IO_ERROR_NONE && errno != EINPROGRESS)
    {
      z_io_connect_destroy(self);
      return NULL;
    }
  g_mutex_lock(self->lock);
  self->watch = z_io_watch_fd_new(self->sock, G_IO_OUT, z_io_connect_connected, self);
  g_mutex_unlock(self->lock);
  return self;
}

/*+
 
  Public function which can be called to free up a connection
  instance. 

  Parameters:
    self      instance pointer

  Remarks: It can be called both prior to, and after the connection is
           established. It frees up all associated resources.

  Returns:
    none

  +*/
void 
z_io_connect_destroy(ZIOConnect *self)
{
  g_mutex_lock(self->lock);
  if (self->sock != -1)
    close(self->sock);
  if (self->watch)
    z_io_watch_fd_destroy(self->watch);
  z_sockaddr_unref(self->local);
  z_sockaddr_unref(self->remote);
  g_mutex_unlock(self->lock);
  g_mutex_free(self->lock);
  
  /* if anybody is locked on self->lock, it's a problem */
  g_free(self);
}

/*+
  
  This is an implementation of the public ZIOListen interface with
  some extra hidden attributes. It binds to the given local address,
  and waits for incoming connections. For each arriving connection a
  callback is called.

  +*/
typedef struct _ZIORealListen 
{
  ZSockAddr *local;
  ZIOWatchFD *watch;
  int sock;
  ZAcceptFunc callback;
  gpointer user_data;
  gboolean accept_one;
  GMutex *lock;
} ZIORealListen;

/*+ Private connection passed to a ZIOWatchFD to be called when the
  socket becomes readable.
  +*/
static gboolean z_io_listen_accept(int fd,
                                   GIOCondition condition,
                                   gpointer data)
{
  ZIORealListen *l = (ZIORealListen *) data;
  int newfd;
  ZSockAddr *client;

  if (z_accept(fd, &newfd, &client) == G_IO_ERROR_NONE)
    {
      z_fd_set_nonblock(newfd, 0);
      l->callback(newfd, client, l->accept_one, l->user_data);
      if (l->accept_one)
        {
          z_io_listen_destroy((ZIOListen *) l);
          return FALSE;
        }
    }
  else
    {
      z_warning(0, "Error accepting connection... %m");
    }
  return TRUE;
}

/*+

  This function creates a new ZIOListen instance, internally it uses
  the ZIORealListen structure to represent it's private data.

  Parameters:
    local            address to bind to
    callback         function to call, when an incoming connection is established
    user_data        opaque pointer passed to callback
    accept_one       allow only a single connection to be accepted

  Returns:
    The newly created instance.

  +*/
ZIOListen *z_io_listen_new(ZSockAddr *local, 
			   gboolean accept_one,
			   ZAcceptFunc callback,
			   gpointer user_data)
{
  ZIORealListen *self = g_new0(ZIORealListen, 1);
  char sabuf[1500];
  socklen_t sabuflen = sizeof(sabuf);

  self->callback = callback;
  self->user_data = user_data;
  self->accept_one = accept_one;
  self->lock = g_mutex_new();
  self->sock = socket(local->sa.sa_family, SOCK_STREAM, 0);
  if (self->sock < 0)
    {
      z_io_listen_destroy((ZIOListen *) self);
      return NULL;
    }
  z_fd_set_nonblock(self->sock, 1);
  if (local && z_bind(self->sock, local) != G_IO_ERROR_NONE)
    { 
      z_io_listen_destroy((ZIOListen *) self);
      return NULL;
    }

  if (getsockname(self->sock, (struct sockaddr *) sabuf, &sabuflen) < 0)
    {
      z_warning(0, "getsockname(): failed due to %s.", g_strerror(errno));
    }
  else
    {
      self->local = z_sockaddr_new((struct sockaddr *) sabuf, sabuflen);
    }
  return (ZIOListen *) self;
}

void
z_io_listen_start(ZIOListen *s)
{
  ZIORealListen *self = (ZIORealListen *) s;
  
  if (listen(self->sock, 5) < 0)
    {
      z_warning(0, "listen(): failed due to: %s.", g_strerror(errno));
      self->callback(-1, NULL, 1, self->user_data);
      z_io_listen_destroy((ZIOListen *) self);
      return;
    }
  g_mutex_lock(self->lock);
  self->watch = z_io_watch_fd_new(self->sock, G_IO_IN, z_io_listen_accept, self);
  g_mutex_unlock(self->lock);
}

/*+
 
  Public function which can be called to free up a ZIORealListen
  instance. 

  Parameters:
    self      instance pointer

  Remarks: It can be called both prior to, and after connections are accepted.

  Returns:
    none

  +*/
void 
z_io_listen_destroy(ZIOListen *l)
{
  ZIORealListen *self = (ZIORealListen *) l;

  g_mutex_lock(self->lock);  
  if (self->sock != -1)
    close(self->sock);
  if (self->watch)
    z_io_watch_fd_destroy(self->watch);
  z_sockaddr_unref(self->local);
  g_mutex_unlock(self->lock);
  g_mutex_free(self->lock);

  /* if somebody is locked on self->lock, it's a problem */
  g_free(self);
}

/*+

  Sets the nonblocking state of the given fd.

  Parameters:
    fd          set nb state of this fd
    enable      enable/disable nonblocking state

  Returns
    TRUE if the operation succeeds.

  +*/
gboolean 
z_fd_set_nonblock(int fd, gboolean enable)
{
  int flags;
  if ((flags = fcntl(fd, F_GETFL)) == -1)
    return 0;
  if (enable)
    flags |= O_NONBLOCK;
  else
    flags &= ~O_NONBLOCK;
    
  return fcntl(fd, F_SETFL, flags) != -1;
}

