/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: log.c,v 1.11 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.7
 * Notes:
 *
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.7
 * Notes:
 *
 *
 ***************************************************************************/

#include <zorp/log.h>
#include <zorp/thread.h>

#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "Zorp"

unsigned int verbose_level = 3;
unsigned int debug_flag = 0;

/*+

  Format and send a debug message with the given verbosity level. 

  Parameters:
    level         the minimal verbosity_level when this message is sent
    format        printf style format specifiers

  Returns: 
    none

  +*/
void 
z_debug(int level, char *format, ...)
{
  va_list l;
  
  if (debug_flag && (level <= verbose_level))
    {
      va_start(l, format);
      g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, format, l);
      va_end(l);
    }
}

/*+

  Format and send a simple message with the given verbosity level. 

  Parameters:
    level         the minimal verbosity_level when this message is sent
    format        printf style format specifiers

  Returns: 
    none

  +*/
void
z_message(int level, char *format, ...)
{
  va_list l;

  if (level <= verbose_level)
    {
      va_start(l, format);
      g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, format, l);
      va_end(l);
    }
}

/*+

  Format and send a warning message with the given verbosity level. 

  Parameters:
    level         the minimal verbosity_level when this message is sent
    format        printf style format specifiers

  Returns: 
    none

  +*/
void 
z_warning(int level, char *format, ...)
{
  va_list l;

  if (level <= verbose_level)
    {
      va_start(l, format);
      g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, format, l);
      va_end(l);
    }
}

/*+

  Format and send a fatal error message with the given verbosity level. 

  Parameters:
    level         unused, just to conform to the rest of z_* log functions
    format        printf style format specifiers

  Returns: 
    none

  +*/
void 
z_fatal(int level, char *format, ...)
{
  va_list l;

  va_start(l, format);
  g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, format, l);
  va_end(l);
}

/*+
  
  Used by the Python interface to send logs with the given verbosity
  level and priority.

  Parameters:
    level        verbosity level
    log_flags    log priority
    msg          formatted log message

  Returns:
    none

  +*/
void
z_log(int level, GLogLevelFlags log_flags, char *msg)
{
  if (level <= verbose_level)
    {
      g_log(G_LOG_DOMAIN, log_flags, "%s", msg);
    }
}

/*+ log handler function to send messages to syslog +*/
static void
z_log_func(const gchar *log_domain,
	   GLogLevelFlags log_flags,
	   const gchar *message,
	   gpointer user_data)
{
  int pri = LOG_INFO;
  if (log_flags & G_LOG_LEVEL_DEBUG)
    pri = LOG_DEBUG;
  else if (log_flags & G_LOG_LEVEL_WARNING)
    pri = LOG_WARNING;
  else if (log_flags & G_LOG_LEVEL_ERROR)
    pri = LOG_ERR;
  syslog(pri, "%s", message);
}

void
z_fetch_stderr_thread(gpointer arg)
{
  gint fd = (gint) arg;
  FILE *log;
  gchar line[1024];
  
  log = fdopen(fd, "r");
  if (!log)
    z_warning(0, "Error during fdopen(): %m");
  line[1023] = 0;
  while (fgets(line, sizeof(line) - 1, log))
    {
      int len = strlen(line);
      if (len == 0)
        continue;
      if (line[len - 1] == '\n')
        line[len - 1] = 0;
      if (line[0] == 0)
        continue;
      z_warning(0, "%s", line);
    }
}

/*+ 

  Initialize Zorp logging with the given name.

  Parameters:
    syslog_name      name to use in syslog messages
    use_syslog       enable sending messages to syslog

  Returns
    none

  +*/
void
z_log_init(gchar *syslog_name, gboolean use_syslog)
{
  if (use_syslog)
    {
      int grab[2];
      
      openlog(syslog_name, LOG_NDELAY | LOG_PID, ZORP_SYSLOG_FACILITY);
      g_log_set_handler(G_LOG_DOMAIN, 0xff, z_log_func, NULL);

      if (pipe(grab) < 0)
        {
          z_warning(0, "Error creating stderr-syslog pipe.");
        }
      z_thread_new("stderr log thread", z_fetch_stderr_thread, (gpointer) grab[0]);

      dup2(grab[1], 1);
      dup2(grab[1], 2);
      close(grab[1]);
    }
}
