/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pyconnect.c,v 1.20 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.16
 * Notes:
 *   - comments
 *   - no process identifier is messages
 *   - z_py_zorp_connect_new_instance() line 150
 *       z_debug ' ? : ' exps: "NULL" and NULL ...
 *
 ***************************************************************************/

#include <zorp/pyconnect.h>
#include <zorp/io.h>
#include <zorp/log.h>
#include <zorp/policy.h>
#include <zorp/pysockaddr.h>

/*
 * struct ZorpConnect
 *
 * Author:  Bazsi, 2000/03/27
 * Purpose: this class encapsulates a connector
 *
 */

struct _ZorpConnect
{
  PyObject_HEAD
  ZIOConnect *connect;
  PyObject *handler;
  int fd;
  GMutex *lock;
  GCond *connected;
};

static PyObject *z_py_zorp_connect_new_instance(PyObject *s, PyObject *args);
static void z_py_zorp_connect_destroy(ZorpConnect *self);
static void z_py_zorp_connect_free(ZorpConnect *self);
static PyObject *z_py_zorp_connect_getattr(PyObject *o, char *name);

static PyObject *z_py_zorp_connect_destroy_method(ZorpConnect *self, PyObject *args);


PyMethodDef z_py_zorp_connect_funcs[] =
{
  { "Connect",  z_py_zorp_connect_new_instance, METH_VARARGS },
  { NULL,      NULL }   /* sentinel*/
};

static PyMethodDef py_zorp_connect_methods[] =
{
  { "destroy",     (PyCFunction) z_py_zorp_connect_destroy_method, 0 },
  { NULL,          NULL }   /* sentinel*/
};

PyTypeObject z_py_zorp_connect_type = 
{
  PyObject_HEAD_INIT(&z_py_zorp_connect_type)
  0,
  "ZorpConnect",
  sizeof(ZorpConnect),
  0,
  (destructor) z_py_zorp_connect_free,
  0,
  (getattrfunc) z_py_zorp_connect_getattr,
  0,
  0,
  0, /*(reprfunc) z_py_zorp_connect_repr,*/
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  "ZorpConnect class for Zorp",
};

static void
z_py_zorp_connect_connected(int fd, gpointer user_data)
{
  ZorpConnect *self = (ZorpConnect *) user_data;
  PyObject *res;

  if (self->handler != Py_None)
    {
      z_policy_lock(z_python_global_state);
      res = PyEval_CallFunction(self->handler, "(i)", fd);
      if (!res || PyErr_Occurred())
	{
	  PyErr_Print();
	  close(fd);
	}
      self->connect = NULL;
      z_policy_unlock(z_python_global_state);
      z_py_zorp_connect_destroy(self);
    }
  else
    {
      g_mutex_lock(self->lock);
      self->connect = NULL;
      self->fd = fd;
      g_mutex_unlock(self->lock);
      g_cond_signal(self->connected);
    }
}

static PyObject *
z_py_zorp_connect_new_instance(PyObject *s, PyObject *args)
{
  ZorpSockAddr *local, *remote;
  ZorpConnect *self;
  PyObject *handler;
  char buf1[128], buf2[128];

  /* called by python, no need to lock */
   
  if (!PyArg_ParseTuple(args, "OOO", &local, &remote, &handler))
    return NULL;
  if (handler != Py_None && !PyCallable_Check(handler))
    {
      PyErr_SetString(PyExc_TypeError, "Handler parameter must be callable");
      return NULL;
    }
  if ((((PyObject *) local != Py_None) && !z_py_zorp_sockaddr_check(local)) ||
      (((PyObject *) remote != Py_None) && !z_py_zorp_sockaddr_check(remote)))
    {
      PyErr_SetString(PyExc_TypeError, "Local and remote arguments must be ZorpSockAddr or None");
      return NULL;
    }
  z_debug(10, "z_py_zorp_connect_new_instance(), local=%s, remote=%s", 
	  (PyObject *) local != Py_None ? z_sockaddr_format(local->sa, buf1, sizeof(buf1)) : "NULL",
	  (PyObject *) remote != Py_None ? z_sockaddr_format(remote->sa, buf2, sizeof(buf2)) : "NULL");

  self = PyObject_NEW(ZorpConnect, &z_py_zorp_connect_type);
  if (!self)
    return NULL;
  
  if (handler == Py_None)  
    {
      self->connected = g_cond_new();
      self->lock = g_mutex_new();
      g_mutex_lock(self->lock);
    }
  else
    {
      self->connected = NULL;
      self->lock = NULL;
    }

  self->handler = handler;
  Py_XINCREF(handler);

  self->connect = z_io_connect_new((PyObject *) local == Py_None ? NULL : local->sa, remote->sa, z_py_zorp_connect_connected, self);
  if (!self->connect)
    {
      PyErr_SetString(PyExc_IOError, "Error during connect");
      
      if (self->lock)
        g_mutex_unlock(self->lock);
        
      Py_XDECREF(handler);
      self->handler = NULL;
      Py_XDECREF(self);
      return NULL;
    }

  if (handler != Py_None)  
    {
      Py_XINCREF((PyObject *) self); /* returned reference */
      return (PyObject *) self;
    }
  else
    {
      PyObject *res;

      Py_BEGIN_ALLOW_THREADS
      g_cond_wait(self->connected, self->lock);
      Py_END_ALLOW_THREADS
      
      g_mutex_unlock(self->lock);

      res = Py_BuildValue("i", self->fd);
      Py_XDECREF((PyObject *) self);
     
      return res;
    }
}

/* called automatically to free up this instance */
static void
z_py_zorp_connect_free(ZorpConnect *self)
{
  if (self->handler)
    {
      Py_XDECREF(self->handler);
      self->handler = NULL;
    }
  if (self->connect)
    {
      z_io_connect_destroy(self->connect);
      self->connect = NULL;
    }
  if (self->lock)
    {
      g_mutex_free(self->lock);
    }
  if (self->connected)
    {
      g_cond_free(self->connected);
    }
  PyMem_DEL(self);
}

/* called from C to destroy this instance */
static void
z_py_zorp_connect_destroy(ZorpConnect *self)
{
  z_policy_lock(z_python_global_state);
  Py_XDECREF(self->handler); /* avoid circular references */
  self->handler = NULL;
  Py_XDECREF((PyObject *) self);
  z_policy_unlock(z_python_global_state);
}

/* called from Python to destroy this instance */
static PyObject *
z_py_zorp_connect_destroy_method(ZorpConnect *self, PyObject *args)
{
  Py_XDECREF(self->handler); /* avoid circular references */
  self->handler = NULL;
  if (self->connect)
    {
      Py_XDECREF(self);
    }
  Py_XINCREF(Py_None);
  return Py_None;
}

static PyObject *
z_py_zorp_connect_getattr(PyObject *o, char *name)
{
  return Py_FindMethod(py_zorp_connect_methods, o, name);
}

void
z_py_zorp_connect_init(void)
{
  PyImport_AddModule("Zorp.Zorp");
  Py_InitModule("Zorp.Zorp", z_py_zorp_connect_funcs);
}
