/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: satyr.c,v 1.11 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Buci
 * Auditor : kisza
 * Last version: 1.6
 * Notes:
 *   - process identifiers in messages (Satyr will be a proxy!)
 * 
 ***************************************************************************/

#include <zorp/satyr.h>
#include <zorp/authconn.h>
#include <zorp/labelset.h>
#include <zorp/log.h>
#include <zorp/lineio.h>
#include <zorp/giossl.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include <string.h>

char *z_satyr_privkey_file = ZORP_SYSCONFDIR "/zorp.key";
char *z_satyr_cert_file = ZORP_SYSCONFDIR "/zorp.crt";
char *z_satyr_ca_file = ZORP_SYSCONFDIR "/ca.crt";

typedef struct _ZRealSatyrAuthenticator
{
  guint ref_count;
  ZAuthConnection *auth_conn;
  SSL_CTX *ssl_ctx;
} ZRealSatyrAuthenticator;

/*
 *  Creates a new ZSatyrAuthenticator, and sets up a new SSL contex for it.
 *  On error it returns NULL. Else the pointer to the newly allocated
 *  ZSatyrAuthenticator.
 */
ZSatyrAuthenticator *
z_satyr_authenticator_new(ZAuthConnection *auth_conn)
{
  ZRealSatyrAuthenticator *self;
  SSL_METHOD *ssl_meth;
  SSL_CTX *ctx;

  GString *error;

  g_return_val_if_fail (auth_conn, NULL);

  self = g_new0 (ZRealSatyrAuthenticator, 1);

  if (!self)
    return NULL;

  self->auth_conn = auth_conn;

#if 0
  SSL_load_error_strings ();
  OpenSSL_add_all_algorithms();

  ssl_meth = SSLv3_server_method ();
  self->ssl_ctx = ctx = SSL_CTX_new (ssl_meth);

  if (!ctx)
    {
      error=g_string_new(NULL);
      ERR_error_gstring(error);
      z_warning (0, "z_satyr: Error creating SSL context: %s",error->str);
      g_string_free(error,0);
      g_free (self);
      return NULL;
    }

  if (SSL_CTX_use_PrivateKey_file
      (ctx, z_satyr_privkey_file, SSL_FILETYPE_PEM) < 0
      || SSL_CTX_use_certificate_file (ctx, z_satyr_cert_file,
				       SSL_FILETYPE_PEM) < 0
      || !SSL_CTX_check_private_key (ctx))
    {
      error=g_string_new(NULL);
      ERR_error_gstring(error);
      z_warning (0, "Error loading keys: %s",error->str);
      g_string_free(error,0);
      SSL_CTX_free (ctx);
      g_free (self);
      return NULL;
    }

  SSL_CTX_set_verify (ctx, SSL_VERIFY_NONE, NULL);
#endif
  self->ref_count = 1;

  return (ZSatyrAuthenticator *) self;
}

/*
 *  destroys the ZSatyrAuthenticator.
 *  It is invoked by z_satyr_authenticator_unref when needed.
 */
static void
z_satyr_authenticator_destroy(ZSatyrAuthenticator *s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail (s);

#if 0
  SSL_CTX_free (self->ssl_ctx);
#endif
  g_free (self);
}

/*
 *  increases the ref_num in the ZSatyrAuthenticator
 */
void
z_satyr_authenticator_ref(ZSatyrAuthenticator *s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail (s);

  self->ref_count++;
}

/*
 *  Decreases the ref_num in a ZSatyrAuthenticator, and destroys it, if needed.
 */
void
z_satyr_authenticator_unref(ZSatyrAuthenticator * s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail(s);

  self->ref_count--;
  if (self->ref_count == 0)
    z_satyr_authenticator_destroy(s);
}

/*
 *  Creates a connection to the given address on the port defined
 *  by 'port'. Connection type (IPv4 or IPv6 ...) depends
 *  on the type of 'client'.
 *  Returns the connected fd, or -1 on error.
 */
static int
z_satyr_make_connection(ZSockAddr *client, int port)
{
  int sock, ret;
  ZSockAddr *sn;
  
  sn = z_sockaddr_new(&client->sa, client->salen);
  switch (sn->sa.sa_family)
    {
    case AF_INET:
      ((struct sockaddr_in *) &sn->sa)->sin_port = htons(port);
      break;
#if 0
    case AF_INET6:
      break;
#endif
    default:
      return -1;
    }

#if 0
  sn.sin_family = AF_INET;
  sn.sin_addr.s_addr = htonl(ip);
  sn.sin_port = htons(SATYR_PORT);
#endif

  sock = socket(sn->sa.sa_family, SOCK_STREAM, 0);
  if (sock < 0)
    {
      z_warning(3, "z_satyr: Cannot create socket %s", strerror (errno));
      z_sockaddr_unref(sn);
      return -1;
    }

  ret = z_connect(sock, sn);
  if (ret != G_IO_ERROR_NONE)
    {
      z_warning(3, "z_satyr: Cannot connect %s", strerror(errno));
      z_sockaddr_unref(sn);
      return -1;
    }

  z_sockaddr_unref(sn);
  return sock;
}

/*
 *  Sends the first, authenticatee selector command to the satyr client,
 *  (on unix satyrmpxd) Returns G_IO_ERROR_NONE on succes, and
 *  G_IO_ERROR_UNKNOWN or error.
 */
static int
z_satyr_choose_user(int fd, guint type, ZSockAddr *addr)
{
  char buffer[16];

  g_assert(addr->sa.sa_family == AF_INET);
  
  g_snprintf(buffer, 15, "%s/%05d\n",
	     (type == Z_SATYR_TYPE_UDP ? "UDP" : "TCP"), 
	     ntohs(((struct sockaddr_in *) &addr->sa)->sin_port));
  write(fd, buffer, 10);
  read(fd, buffer, 4);
  buffer[4] = 0;
  if (!strncmp(buffer, "ACK\n", 4))
    return G_IO_ERROR_NONE;

  return G_IO_ERROR_UNKNOWN;
}

/*
 *  Creates a new SSL GIOChannel on 'fd', useing 'ctx' as the SSL context.
 *  Returns the newly allocated GIOChannel structure, or NULL on error.
 */
static GIOChannel *
z_satyr_make_ssl_connection (SSL_CTX * ctx, int fd)
{
  SSL *ssl;
  GString *error;
  
#if 0
  g_return_val_if_fail (ctx, NULL);
  ssl = SSL_new (ctx);
  if (!ssl)
    {
      error=g_string_new(NULL);
      ERR_error_gstring(error);
      z_warning (0, "Error creating SSL structure: %s",error);
      g_string_free(error,0);
      return NULL;
    }

  SSL_set_fd (ssl, fd);

  if (SSL_accept (ssl) < 0)
    {
      error=g_string_new(NULL);
      ERR_error_gstring(error);
      z_warning (0, "Error accepting SSL connection: %s",error);
      g_string_free(error,0);
      SSL_free (ssl);
      return NULL;
    }
  return g_io_channel_ssl_new (ssl);
#endif
  return g_io_channel_unix_new(fd);
}

/*
 *  Theese are the states of the authenticator state-machine
 */
typedef enum
{
  Z_SATYR_STATE_TRAP=0,
  Z_SATYR_STATE_INIT,
  Z_SATYR_STATE_BASE,
  Z_SATYR_STATE_USERNAME,
  Z_SATYR_STATE_METHOD,
  Z_SATYR_STATE_RESPONSE
} SatyrStates;

/*
 *  The followings are command sets, which says which commands may be used
 *  in a state of the protocol.
 */
static Command user_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {NULL, 0}
};

static Command meth_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {"CHME", 1},
  {NULL, 0}
};

static Command resp_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {"CHME", 1},
  {"RESP", 1},
  {NULL, 0}
};

typedef enum
{
  Z_SATYR_CMD_NOAUTH=0,
  Z_SATYR_CMD_USER,
  Z_SATYR_CMD_CHME,
  Z_SATYR_CMD_RESP
}SatyrCommands;

/*
 *  Calculates which methods may be used. It takes a list of useable methods
 *  in (*methods), which are defined by the policy. It also takes a username
 *  and asks the authentication server for the available methods for the 
 *  specific user. Then takes the intercept of the two lists.
 */
static void
z_satyr_calculate_methods (ZSatyrAuthenticator * s,
			   GList ** methods, char *username)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  char **usrmethods;
  int methnum, i;
  GList *p;
  GList *old_method_list = *methods;
  GList *new_method_list = NULL;

  usrmethods = z_authconn_utyp (self->auth_conn, username, &methnum);
  if (!usrmethods)
    return;

  for (i = 0; i < methnum; i++)
    {
      p = old_method_list;
      while (p)
	{
	  if (!strcmp (p->data, usrmethods[i]))
	    {
	      new_method_list = g_list_append (new_method_list, p->data);
	      old_method_list = g_list_remove (old_method_list, p->data);
	      break;
	    }
	  p = p->next;
	}
    }

  g_list_free (old_method_list);

  *methods = new_method_list;

  return;
}

/*
 *  Gets a challange for user 'username' for the authentication method named
 *  in 'method' from the authentication server. Returns it as a newly allocated
 *  string which should be freed later by the caller by a g_free.
 *  If no challange is available it return NULL.
 */
static char *
z_satyr_get_challange(ZSatyrAuthenticator * s,
		      char *method, char *username)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  char buffer[1024];
  char *ret;

  if(z_authconn_chal (self->auth_conn, method, username, buffer, 1023))
    {
      ret = g_new0(char, strlen(buffer)+1);
      strcpy(ret,buffer);
    }
  else
    {
      ret = NULL;
    }

  return ret;
}


/*
 *  Authenticates the user using the given parameters to the authentication
 *  server. Returns exactly what the auth_server returns.
 */
static ZLabelSet *
z_satyr_real_authenticate(ZSatyrAuthenticator * s,
			  char *username,
			  char *method,
			  char *chal, char *response)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  
  return z_authconn_auth (self->auth_conn, method, username, chal, response);
}


/*
 *  Processes a complete authentication procedure. Connects to the client,
 *  chooses the appropriate user, and talks to satyrd useing the correct
 *  protocol.
 */
ZLabelSet *
z_satyr_authenticator_auth(ZSatyrAuthenticator *s,
			   ZSatyrSubjID *subj,
			   int satyr_port,
			   char *object_descr, char **methods)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  int fd, ret;
  GIOChannel *channel;
  GList *usable_methods = NULL;
  ZLabelSet *return_value = NULL;
  char **p;
  
// state-machine's variables
  char *username = NULL, *method = NULL, *chal = NULL;
  Command *cmd_set;		// available commands
  SatyrStates state;
  int retry;
  
  GList *line;
  SatyrCommands which;
  int argnum;

  g_return_val_if_fail (s, NULL);
  g_return_val_if_fail (subj, NULL);
  g_return_val_if_fail (object_descr, NULL);
  g_return_val_if_fail (methods, NULL);

  fd = z_satyr_make_connection(subj->addr, satyr_port);
  if (fd < 0)
    return NULL;

  ret = z_satyr_choose_user(fd, subj->type, subj->addr);
  if (ret != G_IO_ERROR_NONE)
    {
      close (fd);
      return NULL;
    }

  channel = z_satyr_make_ssl_connection (self->ssl_ctx, fd);
  if (!channel)
    {
      close (fd);
      return NULL;
    }

  state = Z_SATYR_STATE_INIT;
  cmd_set = user_cmds;
  retry = 2;
  while (state != Z_SATYR_STATE_TRAP)
    {
      switch (state)
	{
	case Z_SATYR_STATE_INIT:
	  write_token_line (channel, "ss", "AUTH", object_descr);
	  state = Z_SATYR_STATE_BASE;
	  break;
	case Z_SATYR_STATE_BASE:
	  line = read_command (channel, cmd_set, (int *) &which, &argnum);
	  if (!line)
	    {
	      z_message(5, "z_satyr: Satyr peer failed using the protocol");
	      state = Z_SATYR_STATE_TRAP;
	      break;
	    }
	  switch (which)
	    {
	    case Z_SATYR_CMD_NOAUTH:
	      state = Z_SATYR_STATE_TRAP;
	      break;
	    case Z_SATYR_CMD_USER:
	      state = Z_SATYR_STATE_USERNAME;
	      break;
	    case Z_SATYR_CMD_CHME:
	      state = Z_SATYR_STATE_METHOD;
	      break;
	    case Z_SATYR_CMD_RESP:
	      state = Z_SATYR_STATE_RESPONSE;
	      break;
	    default:
	      z_warning (5, "z_satyr: there is some error...");
	      state = Z_SATYR_STATE_TRAP;
	      break;
	    }
	  break;
	case Z_SATYR_STATE_USERNAME:

	  if (username)
	    g_free (username);

	  username = g_new0(char, strlen (line->next->data) + 1);
	  strcpy(username, line->next->data);
	  free_g_list_of_strings(line);

	  p = methods;
	  while (*p)
	    {
	      usable_methods = g_list_append(usable_methods, *p);
	      p++;
	    }
	  
	  z_satyr_calculate_methods(s, &usable_methods, username);
	  if (!usable_methods)
	    write_token_line(channel, "s", "METH");
	  else
	    write_token_line(channel, "sg", "METH", usable_methods);

	  cmd_set = meth_cmds;
	  state = Z_SATYR_STATE_BASE;
	  break;

	case Z_SATYR_STATE_METHOD:

	  if (method)
	    g_free (method);

	  method = g_new0 (char, strlen (line->next->data) + 1);
	  strcpy (method, line->next->data);
	  free_g_list_of_strings (line);

	  if (chal)
	    g_free (chal);

	  chal = z_satyr_get_challange(s, method, username);
    
          if (chal)
            {
	      write_token_line(channel, "ss", "CHAL", chal);
   	      cmd_set = resp_cmds;
            }
          else
            write_token_line(channel,"s","CHAL");

	  state = Z_SATYR_STATE_BASE;
	  break;

	case Z_SATYR_STATE_RESPONSE:
	  return_value =
	    z_satyr_real_authenticate (s, username, method, chal,
				       line->next->data);
	  free_g_list_of_strings (line);

	  if (return_value)
	    write_token_line (channel, "s", "ACCEPT");
	  else
	    {
	      if (retry--)
		{
		  write_token_line(channel, "s", "RETRY");
		  state = Z_SATYR_STATE_BASE;
		  break;
		}
	      else
	    	write_token_line (channel, "s", "REJECT");
	    }

	  state = Z_SATYR_STATE_TRAP;
	  break;
	case Z_SATYR_STATE_TRAP:
	  break;
	default:
	  z_warning (5, "z_satyr_authenticator: unknown state");
	  break;
	}
    }

  if (username)
    g_free(username);

  if (method)
    g_free(method);

  if (chal)
    g_free(chal);

  g_io_channel_close(channel);
  g_io_channel_unref(channel);
  return return_value;
}
